__VERSION_MAJOR__ = '1'
__VERSION_MINOR__ = '0'
__VERSION_BUILD__ = '1'
__VERSION_TAGS__  = ''

__VERSION_LIST__ = [
    __VERSION_MAJOR__,
    __VERSION_MINOR__,
    __VERSION_BUILD__,
    __VERSION_TAGS__]

__VERSION__ = f'{__VERSION_MAJOR__}.{__VERSION_MINOR__}.{__VERSION_BUILD__}{__VERSION_TAGS__}'

__DATA_VERSION__ = 1

from audata.file import File



import pandas as pd
import h5py as h5


class Annotation:
    def __init__(self, root):
        if not isinstance(root, h5.HLObject):
            raise Exception('Invalid  HDF5 high-level object.')

        self._root = root

    @property
    def lines(self):
        if 'line' in self._root.attrs:
            return pd.DataFrame(data=self._root.attrs['line'])
        else:
            return pd.DataFrame()

    @property
    def vranges(self):
        if 'vrange' in self._root.attrs:
            return pd.DataFrame(data=self._root.attrs['vrange'])
        else:
            return None

    @property
    def tranges(self):
        if 'trange' in self._root.attrs:
            return pd.DataFrame(data=self._root.attrs['trange'])
        else:
            return None

    def add_time_range(self, t0, t1):
        df = self.tranges



import os
import argparse
from glob import glob
import pandas as pd
from dateutil.parser import parse, ParserError
import h5py as h5
import datetime as dt

from typing import Optional

import audata


def _mkfn(path : str) -> str:
    if os.path.isdir(path):
        return os.path.basename(path) + '.h5'
    else:
        return os.path.splitext(os.path.basename(path))[0] + '.h5'

def _addcsv(f : audata.File, nm : str, path : str):
    df = pd.read_csv(path)
    for col in df.columns:
        # Does it have a name that can be interpreted as a "numeric" time?
        is_datetime = col.lower() in ('time', 'timestamp', 'index')
        is_str = df[col].dtype in (str, h5.string_dtype())

        if is_datetime or is_str:
            # Can we convert to date?
            if is_str:
                try:
                    _ = parse(df[col].values[0])
                    is_datetime = True
                except ParserError:
                    pass

            if is_datetime:
                if df[col].dtype in (str, h5.string_dtype()):
                    print('    Parsing time string column: {}'.format(col))
                    df[col] = df[col].apply(parse)
                else:
                    print('    Parsing time offset column: {}'.format(col))
                    df[col] = f.time_reference + df[col]*dt.timedelta(seconds=1)

            else:
                arity = len(df[col].unique())
                pct = float(arity)/len(df)*100
                if pct > 10:
                    print('    Found string column: {} (arity {}%)'.format(col, round(pct)))
                else:
                    print('    Parsing categorical column: {} (arity {} / {}%)'.format(
                        col, arity, round(pct)))
                    df[col] = pd.Series(df[col], dtype='category')
        else:
            print('    Found {} column: {}'.format(df[col].dtype, col))
    f[nm] = df

def _walk(f : audata.File, path : str, prefix : Optional[str] = None):
    if path is None or len(path) == 0 or path[0] == '.':
        return
    elif os.path.isdir(path):
        prefix = '' if prefix is None else os.path.basename(path)
        for next_path in glob('{}/*'.format(path)):
            _walk(f, next_path, prefix)
    elif os.path.splitext(path)[1].lower() == '.csv':
        if prefix is None:
            prefix = ''
        nm = '{}{}{}'.format(
            prefix, '/' if prefix != '' else '',
            os.path.splitext(os.path.basename(path))[0])
        print('  adding {}'.format(nm))
        _addcsv(f, nm, path)
    else:
        return

def main():
    """
    Converts CSV files or nested directories containing multiple CSVs to an audata HDF5 file.

    This tool takes in the path to either a single CSV file or a directory to be recursed for all
    CSV files. For each, it attempts to intelligently convert each CSV into an `audata` dataset
    and store the output to a file (with the same name as the CSV or directory, ending in '.h5').

    The tool will infer a datetime column if it is named time, timestamp, or index, or if it is a
    string column whose first element can be parsed as a datetime object using `dateutils.parser.parse`.
    This parsing does rely on the first row being valid.

    Other string columns will be treated either as factor/categorical columns if the arity (unique values)
    is less than 10% of the total number of rows, otherwise a string column.

    Args:
        path (str): Path to a CSV file or directory to be recursively scanned for CSVs.
    """
    parser = argparse.ArgumentParser(
        description='Converts CSV files or nested directories containing multiple CSVs to an audata HDF5 file.')
    parser.add_argument('path', type=str, help='Path to a CSV file or directory to be recursively scanned for CSVs.')
    args = parser.parse_args()

    fn = _mkfn(args.path)
    print('Creating {}'.format(fn))
    with audata.File.new(fn, overwrite=True) as f:
        _walk(f, args.path)
        print(f)

if __name__ == '__main__':
    main()



import numpy as np
import pandas as pd
import h5py as h5

from typing import Union, AbstractSet, Optional, Tuple, Any, Dict

from audata import _utils as utils
from audata.element import Element


class Dataset(Element):
    """
    Maps to an HDF5 dataset, maintaining the `audata` schema and facilitating translation of
    higher-level data types. Generally should not be instantiated directly.
    """
    def __init__(self,
            au_parent : Element,
            name : str):
        if not isinstance(au_parent, Element):
            raise Exception(f'Invalid parent: {type(au_parent)}')

        parent = au_parent._h5
        if not isinstance(parent, h5.Group):
            raise Exception(f'Invalid parent: {type(parent)}')

        if name not in parent or not isinstance(parent[name], h5.Dataset):
            raise Exception(f'Path {name} is not a dataset in {parent.file.filename}:{parent.name}')

        super().__init__(au_parent, name)

    @classmethod
    def new(cls,
            au_parent : Element,
            name : str,
            value : Union[h5.Dataset, np.ndarray, np.recarray, pd.DataFrame],
            overwrite : bool = False,
            **kwargs) -> 'Dataset':

        if not isinstance(au_parent, Element):
            raise Exception('Must send Element.')

        parent = au_parent._h5
        if not isinstance(parent, h5.Group):
            raise Exception(f'Invalid parent: {type(parent)}')

        if name in parent and not overwrite:
            raise Exception(f'{name} already exists.')

        # If given an HDF5 dataset or an Dataset, read in all of its data
        # (HDF5 dataset as a numpy ndarray, Dataset as a pandas DataFrame).
        if isinstance(value, (h5.Dataset, Dataset)):
            value = value[:]

        # Try to create a class now.

        if isinstance(value, (np.ndarray, np.recarray)):
            return cls.__new_from_array(au_parent, name, value, **kwargs)

        elif isinstance(value, pd.DataFrame):
            return cls.__new_from_dataframe(au_parent, name, value)

        else:
            raise Exception(f'Unsure how to convert type {type(value)}')

    @classmethod
    def __new_from_array(cls,
            au_parent : Element,
            name : str,
            arr : Union[np.ndarray, np.recarray],
            time_cols : AbstractSet[str] = {},
            timedelta_cols : AbstractSet[str] = {}) -> 'Dataset':

        # ATW: TODO: Less lame.
        if arr.dtype.names is None:
            return cls.__new_from_dataframe(au_parent, name, pd.DataFrame(data=arr))

        meta, recs = utils.audata_from_arr(arr, time_ref=au_parent.file.time_reference,
            time_cols=time_cols, timedelta_cols=timedelta_cols)
        au_parent._h5.create_dataset(
            name, chunks=True, maxshape=(None,),
            compression='gzip', shuffle=True, fletcher32=True, data=recs)
        au_parent._h5[name].attrs['.meta'] = utils.dict2json(meta)
        d = cls(au_parent, name)
        return d

    @classmethod
    def __new_from_dataframe(cls,
            au_parent : Element,
            name : str,
            df : pd.DataFrame,
            time_cols : AbstractSet[str] = {},
            timedelta_cols : AbstractSet[str] = {}) -> 'Dataset':

        meta, recs = utils.audata_from_df(df, time_ref=au_parent.file.time_reference,
            time_cols=time_cols, timedelta_cols=timedelta_cols)
        au_parent._h5.create_dataset(
            name, chunks=True, maxshape=(None,),
            compression='gzip', shuffle=True, fletcher32=True, data=recs)
        au_parent._h5[name].attrs['.meta'] = utils.dict2json(meta)
        d = cls(au_parent, name)
        return d

    def __getitem__(self, idx = slice(-1)) -> pd.DataFrame:
        return self.get(idx)

    def get(self,
            idx = slice(-1),
            raw : bool = False,
            datetimes : Optional[bool] = None) -> pd.DataFrame:

        rec = self._h5[idx]
        if raw: return rec
        if isinstance(rec, np.void):
            rec = np.array([rec], dtype=rec.dtype)
        meta = self.meta

        if datetimes is None:
            datetimes = self.file.return_datetimes
        df = utils.df_from_audata(rec, meta, self.file.time_reference, idx, datetimes)
        return df

    def append(self,
            data : Union[pd.DataFrame, np.recarray],
            direct : bool = False,
            time_cols : AbstractSet[str] = {},
            timedelta_cols : AbstractSet[str] = {}):
        arr = None
        if isinstance(data, np.recarray):
            arr = data
            if not direct:
                _, arr = utils.audata_from_arr(arr, time_ref=self.time_reference,
                    time_cols=time_cols, timedelta_cols=timedelta_cols)
        elif direct:
            raise ValueError(f'Data must be in a recarray already to use direct append! Instead {type(data)} was sent.')
        elif isinstance(data, pd.DataFrame):
            # TODO: This does not validate categorical levels. So keep them the same!
            _, arr = utils.audata_from_df(data, time_ref=self.time_reference,
                time_cols=time_cols, timedelta_cols=timedelta_cols)

        N_data = len(arr)
        self._h5.resize((self.nrow + N_data,))
        self._h5[-N_data:] = arr

    @property
    def ncol(self) -> int:
        return len(self._h5.dtype)

    @property
    def nrow(self) -> int:
        return len(self._h5)

    @property
    def columns(self) -> Dict[str, Any]:
        return self.meta['columns']

    @property
    def shape(self) -> Tuple[int, int]:
        return (self.nrow(), self.ncol())

    def __repr__(self):
        def trunc(s, N=20):
            if isinstance(s, str):
                if len(s) > N:
                    return s[:(N-3)] + '...'
                else:
                    return s
            else:
                return f'{s}'

        lines = []
        ncol = self.ncol
        nrow = self.nrow
        lines.append(f'{self.name}: Dataset [{nrow} rows x {ncol} cols]')
        cols = self.columns
        for col in cols:
            c = cols[col]
            if c['type'] == 'integer':
                tstr = 'integer ({})'.format('signed' if c['signed'] else 'unsigned')
            elif c['type'] == 'factor':
                nlevels = len(c['levels'])
                lvls = ', '.join([trunc(lvl) for lvl in c['levels'][:min(3,nlevels)]])
                if nlevels > 3:
                    lvls += ', ...'
                tstr = f'factor with {nlevels} levels [{lvls}]'
            else:
                tstr = c['type']
            lines.append(f'  {col}: {tstr}')
        return '\n'.join(lines)

    def __str__(self):
        return self.__repr__()



import h5py as h5

from typing import Dict, List, Iterable, Tuple, Union

from audata.element import Element


class Group(Element):
    """
    Group element. Acts largely like a container for datasets. Generally should not be
    instantiated directly.
    """
    def __init__(self,
            au_parent : Element,
            name : str = ''):

        if not isinstance(au_parent, (Element, h5.File)):
            raise Exception(f'Invalid parent: {type(au_parent)}')

        if isinstance(au_parent, Element):
            parent = au_parent._h5
            if not isinstance(parent, h5.Group):
                raise Exception(f'Invalid parent: {type(parent)}')

            if name == '':
                target = parent
            else:
                if name not in parent:
                    raise Exception(f'Path {name} was not found in {parent.file.filename}:{parent.name}')
                else:
                    target = parent[name]
            if not isinstance(target, h5.Group):
                raise Exception(f'Path "{name}" is not a group in {parent.file.filename}:{parent.name}')

        super().__init__(au_parent, name)

    def list(self) -> Dict[str, List[str]]:
        """List all child attributes, groups, and datasets."""
        attrs = list(self._h5.attrs)
        others = list(self._h5)
        groups = [g for g in others if isinstance(self._h5[g], h5.Group)]
        datasets = [d for d in others if isinstance(self._h5[d], h5.Dataset)]
        return {
            'attributes': attrs,
            'groups': groups,
            'datasets': datasets
        }

    def recurse(self) -> Iterable[Tuple[Element, str]]:
        """
        Recursively find all datasets. Groups and datasets prefixed with a period
        are ignored.

        Returns:
            Iterable (generator) of tuples of (object: Element, name: str).
        """
        names = [n for n in list(self._h5) if not n.startswith('.')]
        for name in names:
            elem = self.__getitem__(name)
            if isinstance(elem, Group):
                for stuple in elem.recurse():
                    yield stuple
            else:
                yield (elem, elem.name)

    def __repr__(self):
        lines = []
        l = self.list()
        if self.name in ('', '/'):
            lines.append('<ROOT>')
        else:
            lines.append(self.name)

        for t in l['attributes']:
            lines.append(f'  [A] {t}')
        for t in l['groups']:
            lines.append(f'  [G] {t}')
        for t in l['datasets']:
            lines.append(f'  [D] {t}')
        return '\n'.join(lines)

    def __str__(self):
        return self.__repr__()

    def __getitem__(self, key : str) -> Union['Dataset', 'Group', None]:
        if self._h5 is None:
            raise Exception('No group opened.')

        if key == '':
            return Group(self, key)

        if key in self._h5:
            if isinstance(self._h5[key], h5.Dataset):
                from audata.dataset import Dataset
                return Dataset(self, key)
            elif isinstance(self._h5[key], h5.Group):
                return Group(self, key)
            else:
                raise Exception('Unsure how to handle class: {}'.format(type(self._h5[key])))
        else:
            return None

    def __setitem__(self,
            key : str,
            value : Union[h5.Dataset, 'np.ndarray', 'np.recarray', 'pd.DataFrame', None],
            overwrite : bool = True,
            **kwargs):

        if self._h5 is None:
            raise Exception('No group opened.')

        if value is None:
            if key in self._h5:
                del self._h5[key]
        else:
            from audata.dataset import Dataset
            Dataset.new(self, key, value, overwrite=overwrite, **kwargs)

    def __contains__(self, key : str) -> bool:
        return self._h5.__contains__(key)

    def new_dataset(self,
            name : str,
            value : Union[h5.Dataset, 'np.ndarray', 'np.recarray', 'pd.DataFrame'],
            **kwargs):

        self.__setitem__(name, value, **kwargs)



import pandas as pd
import numpy as np
import h5py as h5
import datetime as dt
import jsbeautifier as jsb
import json

from numpy.lib.recfunctions import drop_fields

from typing import Optional, Dict, Any, AbstractSet, Tuple, Union


def df_from_audata(
        rec,
        meta : Dict[str, Any],
        time_ref : Optional[dt.datetime] = None,
        idx = slice(-1),
        datetimes : bool = True) -> pd.DataFrame:

    df = pd.DataFrame(data=rec)
    for col in meta['columns']:
        m = meta['columns'][col]
        if m['type'] == 'factor':
            df[col] = pd.Categorical.from_codes(df[col].values, m['levels'])
        elif m['type'] == 'time':
            if time_ref is None:
                raise Exception('Cannot read timestamps without reference!')
            if datetimes:
                df[col] = time_ref + df[col].values * dt.timedelta(seconds=1)
            else:
                df[col] += time_ref.timestamp()
        elif m['type'] == 'timedelta':
            df[col] = df[col].values * dt.timedelta(seconds=1)
    return df

def audata_from_df(
        df : pd.DataFrame,
        time_ref : Optional[dt.datetime] = None,
        time_cols : AbstractSet[str] = {},
        timedelta_cols : AbstractSet[str] = {}) -> Tuple[Dict[str, Any], np.recarray]:

    cols = list(df)
    columns = {}
    dtype_map = {}
    for col in cols:
        m = {}
        d = df[col].dtype

        if d == h5.string_dtype():
            # String d-type has to be set explicitely or HDF5 won't accept it. Default to
            # variable-length strings, although in the future we might want to support fixed-
            # length strings which can be compressed (vlen strings are NOT compressed).
            dtype_map[col] = h5.string_dtype()
            m['type'] = 'string'
        elif d.name == 'category':
            m['type'] = 'factor'
            m['levels'] = list(df[col].cat.categories)
            m['ordered'] = df[col].cat.ordered
            df[col] = df[col].cat.codes
        elif d.kind in ['i', 'u']:
            m['type'] = 'integer'
            m['signed'] = d.kind == 'i'
        elif d.kind == 'M':
            if time_ref is None:
                raise(Exception('Cannot convert timestamps without time reference!'))

            m['type'] = 'time'
            df[col] = (df[col].dt.tz_convert(time_ref.tzinfo) - time_ref).dt.total_seconds()
        elif d.kind == 'm':
            m['type'] = 'timedelta'
            df[col] = df[col].dt.total_seconds()
        elif col in time_cols:
            # Assume offset from reference in appropriate units.
            m['type'] = 'time'
            dtype_map[col] = 'f8'
        elif col in timedelta_cols:
            # Assume delta in appropriate units.
            m['type'] = 'timedelta'
            dtype_map[col] = 'f8'
        else:
            typenames = {
                'b': 'boolean',
                'f': 'real',
                'c': 'complex'
            }
            m['type'] = typenames[d.kind]
        columns[col] = m

    meta = {'columns': columns}
    rec = df.to_records(index=False, column_dtypes=dtype_map)
    return meta, rec

def audata_from_arr(
        arr : Union[np.ndarray, np.recarray],
        time_ref : Optional[dt.datetime] = None,
        time_cols : AbstractSet[str] = {},
        timedelta_cols : AbstractSet[str] = {}) -> Tuple[Dict[str, Any], np.recarray]:

    # TODO: Factor types are not supported in this mode. Also not supported: non-string objects.
    cols = arr.dtype.names
    columns = {}
    for col in cols:
        m = {}
        d = arr.dtype[col]

        if d.kind == 'M':
            if time_ref is None:
                raise(Exception('Cannot convert timestamps without time reference!'))

            # Note: Since numpy datetime64 types are note timezone aware, we must assume the datetimes
            # use the same timezone.
            m['type'] = 'time'
            arr[col] = (arr[col] - np.datetime64(time_ref.replace(tzinfo=None))) / np.timedelta64(1, 's')
        elif d.kind == 'm':
            m['type'] = 'timedelta'
            arr[col] = arr[col] / np.timedelta64(1, 's')
        elif d == h5.string_dtype():
            if h5.check_vlen_dtype(d) != str:
                arr[col] = arr[col].astype(h5.string_dtype())
            m['type'] = 'string'
        elif col in time_cols:
            # Assume offset from reference in appropriate units.
            m['type'] = 'time'
            if d.type != 'f':
                arr[col] = arr[col].astype('f8')
        elif col in timedelta_cols:
            # Assume delta in appropriate units.
            m['type'] = 'timedelta'
            if d.type != 'f':
                arr[col] = arr[col].astype('f8')
        elif d.kind in ['i', 'u']:
            m['type'] = 'integer'
            m['signed'] = d.kind == 'i'
        else:
            typenames = {
                'b': 'boolean',
                'f': 'real',
                'c': 'complex'
            }
            m['type'] = typenames[d.kind]
        columns[col] = m

    meta = {'columns': columns}
    return meta, arr

def json2dict(json_str : str) -> Dict[str, Any]:
    if not isinstance(json_str, str):
        raise Exception(f'Expecting string, found {type(json_str)}')
    return json.loads(json_str)

def dict2json(json_dict : Dict[str, Any], format : bool = True) -> str:
    if not isinstance(json_dict, dict):
        raise Exception(f'Expecting dictionary, found {type(json_dict)}')
    json_str = json.dumps(json_dict)
    return jsb.beautify(json_str) if format else json_str



import h5py as h5
from typing import Optional, Union, Dict, Any

from audata._utils import json2dict, dict2json


class Element:
    """
    Represents an abstract `audata` element (e.g., files, groups, etc..) It should not
    be necessary to interact with this class directly.
    """
    def __init__(self,
            parent : Optional[Union[h5.File, 'Element']] = None,
            name : str = ''):
        if parent is None:
            self.parent = None
            self.file = None
            self._h5 = None
        elif isinstance(parent, h5.File):
            self.parent = self
            self.file = self
            self._h5 = parent
        elif isinstance(parent, Element):
            self.parent = parent
            self.file = parent.file
            self._h5 = parent._h5 if name == '' else parent._h5[name]
        else:
            raise Exception(f'Invalid parent: {type(parent)}.')

    def clear(self):
        self.parent = None
        self.file = None
        self._h5 = None

    @property
    def name(self) -> Optional[str]:
        """Element name (`Optional[str]`, read-only)"""
        return self._h5.name if self._h5 is not None else None

    @property
    def filename(self) -> Optional[str]:
        """Path to file (`Optional[str]`, read-only)"""
        return self._h5.file.filename if self._h5 is not None else None

    @property
    def time_reference(self) -> Optional['dt.datetime']:
        """File time reference (`Optional[dt.datetime]`, read-only)"""
        return self.file.time_reference if self.file is not None else None

    @property
    def valid(self) -> bool:
        """Is element valid? (`bool`, read-only)"""
        return (self.parent is not None) and (self.file is not None) and (self._h5 is not None)

    @property
    def meta(self) -> Dict[str, Any]:
        """Element meta data (HDF5 .meta attribute) (JSON dictionary)"""
        return json2dict(self._h5.attrs['.meta']) \
            if self.valid and ('.meta' in self._h5.attrs) else None

    @meta.setter
    def meta(self, data : Dict[str, Any]):
        if not self.valid:
            raise Exception('Attempting to set meta on invalid element!')
        self._h5.attrs['.meta'] = dict2json(data)

    @property
    def meta_audata(self) -> Dict[str, Any]:
        """File audata meta, '.meta/audata' attribute (JSON dictionary, read-only)"""
        return json2dict(self._h5.file['.meta'].attrs['audata']) if self.valid else None

    @property
    def meta_data(self) -> Dict[str, Any]:
        """File data meta, '.meta/data' attribute (JSON dictionary, read-only)"""
        return json2dict(self._h5.file['.meta'].attrs['data']) if self.valid else None



import pandas as pd
import numpy as np
import lorem
import os
import datetime as dt
import tzlocal

if __name__ == '__main__':
    def ensure(fn):
        fn = 'patient123/{}'.format(fn)
        os.makedirs(os.path.dirname(fn), exist_ok=True)
        return fn

    time_reference = dt.datetime.now(tzlocal.get_localzone())
    P = 60*30 # s

    print('Making low-rate...')
    Fs = 1./5 # Hz
    N = int(Fs*P)
    pd.DataFrame(
        data={
            'time': [time_reference + dt.timedelta(seconds=x/Fs) for x in range(N)],
            'time2': [time_reference + dt.timedelta(seconds=x/Fs + 0.5) for x in range(N)],
            'ints': list(range(N)),
            'floats': [i*1.1 for i in list(range(N))],
            'strings': [lorem.sentence() for x in range(N)],
            'factor': pd.Series([['Cat', 'Dog', 'Liger'][x % 3] for x in range(N)], dtype='category')
        }).to_csv(ensure('some_table.csv'), index=False)

    print('Making 500Hz ECG...')
    Fs = 500 # Hz
    N = int(Fs*P)
    pd.DataFrame(
        data={
            'time': [x/Fs for x in range(N)],
            'II': np.random.uniform(-2, 2, N),
            'III': np.random.uniform(-2, 2, N),
            'IV': np.random.uniform(-2, 2, N)
        }).to_csv(ensure('waveform/ecg.csv'), index=False)

    print('Making 250Hz waveform...')
    Fs = 250 # Hz
    N = int(Fs*P)
    pd.DataFrame(
        data={
            'time': [x/Fs for x in range(N)],
            'ART': np.random.uniform(50, 90, N),
            'PAP': np.random.uniform(10, 15, N),
            'CVP': np.random.uniform(0, 5, N)
        }).to_csv(ensure('waveform/pressures.csv'), index=False)

    print('Making vitals...')
    Fs = 1.5 # Hz
    N = int(Fs*P)
    pd.DataFrame(
        data={
            'time': [x/Fs for x in range(N)],
            'Sys': np.random.uniform(80, 90, N)
        }).to_csv(ensure('vitals/some_device/Sys.csv'), index=False)
    pd.DataFrame(
        data={
            'time': [x/Fs for x in range(N)],
            'MAP': np.random.uniform(70, 80, N)
        }).to_csv(ensure('vitals/some_device/MAP.csv'), index=False)
    pd.DataFrame(
        data={
            'time': [x/Fs for x in range(N)],
            'Dia': np.random.uniform(50, 65, N)
        }).to_csv(ensure('vitals/some_device/Dia.csv'), index=False)



import pandas as pd
import numpy as np

import datetime as dt
from dateutil import parser
import lorem

from audata import File

if __name__ == '__main__':
    f = File.new('test.h5', overwrite=True)

    # Writing a matrix.
    f['mymat'] = np.ones([100, 300])

    # Writing and appending a data frame.
    N = 1000
    df = pd.DataFrame(
        data={
            'time': [f.time_reference + dt.timedelta(seconds=x/250.0) for x in range(N)],
            'time2': [f.time_reference + dt.timedelta(seconds=x/250.0 + 0.5) for x in range(N)],
            'ints': list(range(N)),
            'floats': [i*1.1 for i in list(range(N))],
            'strings': [lorem.sentence() for x in range(N)],
            'factor': pd.Series([['Cat', 'Dog', 'Liger'][x % 3] for x in range(N)], dtype='category')
        }
    )
    f['test/mydf'] = df
    f['test/mydf'].append(df)

    # Annotations - Not yet.
    # f['mydf'].add_range('time', 'time2')

    # Read some data.
    print(f)
    print(f['.meta'])
    print(f.time_reference)
    print(f['test/mydf'])
    print(f['test']['mydf'][:])

    f.close()

    # Open the created file and do the same thing.
    print('Testing openeing file we just created:')
    with File.open('test.h5') as f:
        # Read some data.
        print(f)
        print(f['.meta'])
        print(f.time_reference)
        print(f['test']['mydf'])
        print(f['test/mydf'][:])
        print(f['test/mydf'][-3:])
        print(f['test/mydf'][-3])
        print(list(f.recurse()))



import os
import h5py as h5
from dateutil import parser
import datetime as dt
import tzlocal
import pandas as pd
import numpy as np

from typing import Optional, Union, Literal

from audata import __VERSION__, __DATA_VERSION__
from audata._utils import json2dict, dict2json
from audata.group import Group

class File(Group):
    """
    Wrapper around an HDF5 file.

    The wrapper adds a lot of convenience in handling audata files by automatically
    maintaining the correct underlying data schema while providing an intuitive
    interface and some convenience functions. Datasets can be accessed and updated
    by using a file object as a dictionary, a dictionary of dictionaries, or a dictionary
    where hierarchy is implied by use of the forward-slash as a "directory" delimiter,
    similar to how datasets are accessed using h5py. Data conversions to store higher-level
    data types unsupported natively by HDF5 (e.g., timestamps, ranges, or categorical
    variables) is handled implicitely.

    Generally, files are opened or created using the `open` or `new` class methods,
    respectively, instead of the constructor.

    Example:
        Creating a new file and adding a dataset:

        >>> f = audata.File.new('test.h5', time_reference=dt.datetime(2020, 5, 4, tzinfo=UTC))
        >>> f['data'] = pd.DataFrame(data={
        ...     'time': f.time_reference + dt.timedelta(hours=1)*np.arange(3),
        ...     'a': [1,2,3],
        ...     'b': pd.Categorical(['a', 'b', 'c'])})
        >>> f['data']
        /data: Dataset [3 rows x 3 cols]
          time: time
          a: integer (signed)
          b: factor with 3 levels [a, b, c]
        >>> f['data'][:]
                               time  a  b
        0 2020-05-04 00:00:00+00:00  1  a
        1 2020-05-04 01:00:00+00:00  2  b
        2 2020-05-04 02:00:00+00:00  3  c
    """
    DateTimeFormat = '%Y-%m-%d %H:%M:%S.%f %Z'

    def __init__(self,
            file : h5.File,
            time_reference : Optional[dt.datetime] = None,
            return_datetimes : bool = True):
        """
        Instantiates the File object.

        Generally `new` or `open` will be called, the constructor is not called
        directly.

        Args:
            file: The opened HDF5 file object.
            time_reference: The file-level time reference.
            return_datetimes: True if timestamps should be converted to `dt.datetime` objects,
                False if Unix timestamps (UTC) should be returned instead.
        """
        if not isinstance(file, h5.File):
            raise Exception(f'Invalid file type: {type(file)}')

        self._time_reference = time_reference
        self.return_datetimes = return_datetimes
        super().__init__(file)
        self.time_reference = time_reference

    def __del__(self):
        if self is not None:
            self.close()

    @property
    def time_reference(self) -> dt.datetime:
        """
        The timezone-aware time reference.

        Can be set with either a `dt.datetime` object or a `str` that can be parsed as
        a datetime. If a naive datetime is provided, the local timezone will be inferred.
        """
        return self._time_reference

    @time_reference.setter
    def time_reference(self, new_ref : Union[str, dt.datetime]):
        if not self.valid:
            raise Exception('Attempting to use uninitialized File!')

        # If it's a date/time string, that's fine, but try to parse it first to make sure
        # the format is always consistent.
        if isinstance(new_ref, str):
            new_ref = parser.parse(new_ref)

        # Now only accept datetime objects.
        if isinstance(new_ref, dt.datetime):
            # We need a timezone. If none is given, assume it's local time.
            if new_ref.tzinfo is None:
                new_ref = tzlocal.get_localzone().localize(new_ref)
            self._time_reference = new_ref
            new_ref_str = new_ref.strftime(File.DateTimeFormat)

            data = self.meta_data
            data['time']['origin'] = new_ref_str
            self._h5['.meta'].attrs['data'] = dict2json(data)

    @classmethod
    def new(cls,
            filename : str,
            overwrite : bool = False,
            time_reference : Union[Literal['now'], dt.datetime] = 'now',
            title : Optional[str] = None,
            author : Optional[str] = None,
            organization : Optional[str] = None,
            return_datetimes : bool = True,
            **kwargs) -> 'File':
        """
        Create a new file.

        Args:
            filename: The path of the file to open.
            overwrite: If True, existing files will be truncated. Otherwise, an
                existing file will cause an exception.
            time_reference: The time reference to use, or 'now' to use the time
                of file creation.
            title: A title for the dataset.
            author: The dataset author.
            organization: The dataset organization.
            return_datetimes: If True times will be converted to `dt.datetime` objects,
                otherwise Unix (UTC) timestamps.
            **kwargs: Additional keyword arguments will be passed on to `h5.File`'s constructor.

        Returns:
            The newly opened file object.
        """
        if os.path.exists(filename) and not overwrite:
            raise Exception('File "{}" already exists!'.format(filename))

        if time_reference == 'now':
            time_reference = dt.datetime.now(tz=tzlocal.get_localzone())

        f = h5.File(filename, 'w', **kwargs)
        f.create_group('.meta')
        f['.meta'].attrs['audata'] = dict2json({
            'version': __VERSION__,
            'data_version': __DATA_VERSION__
        })
        f['.meta'].attrs['data'] = dict2json({
            'title': title,
            'author': author,
            'organization': organization,
            'time': {
                'origin': time_reference.strftime(File.DateTimeFormat),
                'units': 'seconds'
            }
        })
        c = cls(f, time_reference=time_reference, return_datetimes=return_datetimes)
        return c

    @classmethod
    def open(cls,
            filename : str,
            create : bool = False,
            readonly : bool = True,
            return_datetimes : bool = True,
            **kwargs) -> 'File':
        """
        Open an audata file.

        Args:
            filename: The path to the file to open.
            create: If True, missing files will be created. Otherwise, missing files
                cause an exception.
            readonly: Whether to open in read-only or mutable.
            return_datetimes: If True times will be converted to `dt.datetime` objects,
                otherwise Unix (UTC) timestamps.
            **kwargs: Additional keyword arguments will be passed on to `h5.File`'s constructor
                if a file is to be created.

        Returns:
            The opened file object.
        """
        if not os.path.exists(filename):
            if create: return cls.new(filename, **kwargs)
            else: raise Exception(f'File not found: {filename}')

        f = h5.File(filename, 'r' if readonly else 'a')
        c = cls(f, return_datetimes=return_datetimes)
        c._time_reference = parser.parse(c.meta_data['time']['origin'])
        return c

    def close(self):
        """Close the file handle."""
        if self._h5 is not None:
            self._h5.close()
            self.clear()

    def flush(self):
        """Flush changes to disk."""
        if self._h5 is not None:
            self._h5.flush()
        else:
            raise Exception('No file opened!')

    def __repr__(self):
        return self.__getitem__('').__repr__()

    def __str__(self):
        return self.__getitem__('').__str__()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()



from setuptools import setup, find_packages
import numpy as np
import os

from audata import __VERSION__

NAME = 'audata'
VERSION = __VERSION__


def read(fn):
    return open(os.path.join(os.path.dirname(__file__), fn)).read()


setup(
    name=NAME,
    version=VERSION,
    description='A python package for reading and writing data for AUView.',
    long_description=read('readme.md'),
    author='Anthony Wertz',
    author_email='awertz@cmu.edu',
    license='GNU LGPL 3',
    entry_points={
        'console_scripts': ['csv2audata=audata.bin.csv2audata:main']
    },
    install_requires=[
        'numpy',
        'pandas',
        'h5py',
        'jsbeautifier',
        'lorem',
        'tzlocal',
        'h5py'
    ],
    python_requires='>=3.8',
    include_dirs=[np.get_include()],
    packages=find_packages()
)
